import pyscan
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
import tqdm
import matplotlib.patches as patches
import time
import seaborn as sns

ds_october = xr.open_dataset(f"data/pm_utah_2016_october.nc")
ds_january = xr.open_dataset(f"data/pm_utah_2016_january.nc")

extents = [ds_october.x.min(), ds_october.x.max(), ds_october.y.min(), ds_october.y.max()]

def upscale_resolution(ds, factor):
    minx, maxx = ds.x.min(), ds.x.max()
    miny, maxy = ds.y.min(), ds.y.max()
    new_x = np.linspace(minx, maxx, len(ds.x) * factor)
    new_y = np.linspace(maxy, miny, len(ds.y) * factor)
    new_ds = ds.interp(x=new_x, y=new_y, method='linear')
    return new_ds

# ds_october = upscale_resolution(ds_october, 4)
# ds_january = upscale_resolution(ds_january, 4)

def xr_to_coords(ds):
    ds_stacked = ds.stack(xy=['y', 'x'])  # Stack x and y into a single dimension
    ds_non_nan = ds_stacked.dropna('xy')  # Drop NaNs along the stacked dimension
    xs = ds.x.values
    x_non_nan = ds_non_nan.x.values
    ys = ds.y.values
    y_non_nan = ds_non_nan.y.values
    values_non_nan = ds_non_nan.value.values
    non_nan_array = np.column_stack((x_non_nan, y_non_nan, values_non_nan))
    return non_nan_array, np.unique(xs), np.unique(ys)


def xr_to_grid1_(ds_m, ds_b):
    print("Net grid...")
    b = []
    m = []
    pts = []
    for i in tqdm.tqdm(range(len(ds_m.x))):
        for j in range(len(ds_m.y)):
            # skip nan values
            # ACCESS is (y, x)
            if np.isnan(ds_m.value[j, i]) or np.isnan(ds_b.value[j, i]):
                continue
            x = ds_m.x[i]
            y = ds_m.y[j]

            v2 = ds_m.value[j, i]
            v1 = ds_b.value[j, i]

            b.append(pyscan.WPoint(v1, x, y, 1)) # january is baseline
            m.append(pyscan.WPoint(v2, x, y, 1)) # october is measure
        
    unique_xs = sorted(list(set(ds_m.x.values)))
    unique_ys = sorted(list(set(ds_m.y.values)))
    zip_ = list(zip(unique_xs, unique_ys))
    pts = [pyscan.Point(x, y, 1) for x, y in zip_]

    grid = pyscan.make_net_grid(pts, m, b)
    return grid


def xr_to_grid1(ds_m, ds_b):
    print("Net grid...")
    b_arr, unique_xs, unique_ys = xr_to_coords(ds_b)
    m_arr, _, _ = xr_to_coords(ds_m)
    pts = [pyscan.Point(x, y, 1) for x, y in zip(unique_xs, unique_ys)]
    b = [pyscan.WPoint(b, x, y, 1.) for b, x, y in zip(b_arr[:, 2], b_arr[:, 0], b_arr[:, 1])]
    m = [pyscan.WPoint(m, x, y, 1.) for m, x, y in zip(m_arr[:, 2], m_arr[:, 0], m_arr[:, 1])]

    grid = pyscan.make_net_grid(pts, m, b)
    return grid


def xr_to_grid0(ds_m, ds_b):
    print("Exact grid...")
    bs = [pyscan.WPoint(ds_b.value[j, i], ds_b.x[i], ds_b.y[j], 1) for i in range(len(ds_b.x)) for j in range(len(ds_b.y)) if not np.isnan(ds_b.value[j, i])]
    ms = [pyscan.WPoint(ds_m.value[j, i], ds_m.x[i], ds_m.y[j], 1) for i in range(len(ds_m.x)) for j in range(len(ds_m.y)) if not np.isnan(ds_m.value[j, i])]

    grid = pyscan.make_exact_grid(ms, bs)
    return grid

def print_grid_info(grid, ds_m, ds_b):
    print('Grid size:', grid.size())
    print('Size should be:', max(len(ds_m.x), len(ds_m.y)))
    print('Grid counts:', len(grid.getRedCounts()))
    print('Red counts sum:', sum(grid.getRedCounts()))
    print('Red counts sum should be:', ds_m.value.values[~np.isnan(ds_m.value.values)].sum())
    print('Blue counts sum:', sum(grid.getBlueCounts()))
    print('Blue counts sum should be:', ds_b.value.values[~np.isnan(ds_b.value.values)].sum())
    
    return grid

def xr_to_grid2(ds_m, ds_b):
    # Quantile based grid, b and m are used to define n (depending on resolution) quantiles
    # then the grid is constructed based on these quantiles.
    print("Resolution grid...")
    # arr ix nx3 x,t,value
    m_arr, unique_xs, unique_ys = xr_to_coords(ds_m)
    b_arr, _, _ = xr_to_coords(ds_b)
    b_points = [pyscan.WPoint(b, x, y, 1.) for b, x, y in zip(b_arr[:, 2], b_arr[:, 0], b_arr[:, 1])]
    m_points = [pyscan.WPoint(w, x, y, 1.) for w, x, y in zip(m_arr[:, 2], m_arr[:, 0], m_arr[:, 1])]
    res = max(len(unique_xs), len(unique_ys))
    grid = pyscan.Grid(res, m_points, b_points)
    return grid

def xr_to_grid3(ds_m, ds_b):
    print("Numpy grid...")
    m_arr, unique_xs, unique_ys = xr_to_coords(ds_m)
    b_arr, _, _ = xr_to_coords(ds_b)
    grid = pyscan.Grid(unique_xs, unique_ys, m_arr, b_arr)
    return grid

def grid_to_np(grid):
    print('Converting grid to numpy...')
    size = grid.size()
    m_arr = np.zeros((size, size))
    b_arr = np.zeros((size, size))
    for y in tqdm.tqdm(range(size)):
        for x in range(size):
            b_arr[y, x] = grid.blueWeight(y, x)  # Note: 
            m_arr[y, x] = grid.redWeight(y, x)
    # remember, in pyscan grid, y,x are sorted from small to large. So we have to flip the y-axis
    m_arr = np.flipud(m_arr)
    b_arr = np.flipud(b_arr)
    return m_arr, b_arr

def plot_result(b_arr, m_arr, extents, rect=None, rect2=None):
    # create 3 subplots, 1 for b one m, one diff. if rect non-None, plot it on the the three subplots
    # use same color scale for all three subplots
    print('Plotting...')

    # Compute global min and max for consistent color scaling
    # Non nans values only
    non_nan_b = b_arr[~np.isnan(b_arr)]
    non_nan_m = m_arr[~np.isnan(m_arr)]
    vmin = min(non_nan_b.min(), non_nan_m.min(), np.abs(non_nan_b - non_nan_m).min())
    vmax = max(non_nan_b.max(), non_nan_m.max(), np.abs(non_nan_b - non_nan_m).max())

    fig, axs = plt.subplots(1, 3, figsize=(9, 3))

    # Plot Baseline (b)
    im1 = axs[0].imshow(b_arr, cmap='coolwarm', vmin=vmin, vmax=vmax, extent=extents)
    # cbar1 = plt.colorbar(im1, ax=axs[0], fraction=0.046, pad=0.04)
    axs[0].set_title('Baseline')

    # Plot Measure (m)
    im2 = axs[1].imshow(m_arr, cmap='coolwarm', vmin=vmin, vmax=vmax, extent=extents)
    # cbar2 = plt.colorbar(im2, ax=axs[0], fraction=0.046, pad=0.04)
    axs[1].set_title('Measure')

    # Plot Absolute Difference |b - m|
    im3 = axs[2].imshow(np.abs(b_arr - m_arr), cmap='coolwarm', extent=extents)
    # cbar3 = plt.colorbar(im3, ax=axs[0], fraction=0.046, pad=0.04)
    axs[2].set_title('Difference')

    # Add rectangle if it exists
    if rect is not None:
        for ax in axs:
            rect_patch = patches.Rectangle(
                (rect.lowX(), rect.lowY()), rect.upX() - rect.lowX(), rect.upY() - rect.lowY(),
                linewidth=2, edgecolor='k', facecolor='none'
            )
            ax.add_patch(rect_patch)
        # add a label to the rectangle indicating its Area Limited Subgrid (450 cells)
        # here
        axs[0].text(rect.lowX() -1, rect.lowY() - 1, 'Area Limited \n Subgrid \n (450 cells)', color='k', fontsize=11)

    
    if rect2 is not None:
        for ax in axs:
            rect_patch = patches.Rectangle(
                (rect2.lowX(), rect2.lowY()), rect2.upX() - rect2.lowX(), rect2.upY() - rect2.lowY(),
                linewidth=2, edgecolor='r', facecolor='none'
            )
            ax.add_patch(rect_patch)
        # add a label to the rectangle indicating its NUnlimited Subgrid
        axs[0].text(rect2.lowX(), rect2.lowY() - 0.5, 'Unlimited Subgrid', color='darkred', fontsize=11)

    plt.tight_layout()
    # plt.show()
    # plt.savefig(f"{RESULTS_PATH}/utah_area_limited.png")
    plt.show()

def run_algo(ds_january, ds_october, extents, grid):
    disc_f = pyscan.RKULLDORF
    print('Running algorithm...')
    subgrid = pyscan.max_subgrid_convex(grid, .0001, disc_f)
    
    col0, col1, row0, row1 = subgrid.lowCol(), subgrid.upCol(), subgrid.lowRow(), subgrid.upRow()
    width = col1 - col0 + 1
    height = row1 - row0 + 1
    fvalue = subgrid.fValue()

    rect = grid.toRectangle(subgrid)
    rect2 = rect

    xB, xU, yB, yU = rect.lowX(), rect.upX(), rect.lowY(), rect.upY()
    # Rect size in number of cells
    rect_size_area = width * height
    print(f'Fvalue: {fvalue}')
    print(f'Width: {width} cells, Height: {height} cells')
    print(f'Area: {rect_size_area} cells')
    print(f'Rectangle: ({col0}, {col1}), ({row0}, {row1})')
    
    plot_result(ds_january.value.values, ds_october.value.values, extents, rect, rect2)
    print('done')


# t0 = time.time()
# grid0 = xr_to_grid0(ds_october, ds_january)
# t1 = time.time()
# print_grid_info(grid0, ds_october, ds_january)
# print(f'Grid 0 (Exact) took {t1 - t0:.2f} seconds')

def test_grid(grid, ds_october, ds_january):

    m_arr = ds_october.value.values
    m_arr = m_arr / np.nansum(m_arr)
    b_arr = ds_january.value.values
    b_arr = b_arr / np.nansum(b_arr)
    m_arr0, b_arr0 = grid_to_np(grid)
    
    # b_arr0[np.isnan(b_arr)] = np.nan

    min_val = np.min([np.nanmin(m_arr), np.nanmin(b_arr), np.nanmin(m_arr0), np.nanmin(b_arr0)])
    max_val = np.max([np.nanmax(m_arr), np.nanmax(b_arr), np.nanmax(m_arr0), np.nanmax(b_arr0)])
    fig, ax = plt.subplots(1, 2, figsize=(10, 5))
    sns.heatmap(m_arr, ax=ax[0], cmap='coolwarm', cbar=False, xticklabels=10, yticklabels=10,
                 vmin=min_val, vmax=max_val)
    ax[0].set_title('Original')
    sns.heatmap(m_arr0, ax=ax[1], cmap='coolwarm', cbar=False, xticklabels=10, yticklabels=10,
                 vmin=min_val, vmax=max_val)
    ax[1].set_title('Recovered')
    plt.show()

    # assert np.allclose(m_arr, m_arr0)
    # assert np.allclose(b_arr, b_arr0)


t0 = time.time()
grid1 = xr_to_grid1(ds_october, ds_january)
t1 = time.time()
print_grid_info(grid1, ds_october, ds_january)
print(f'Grid 1 (Net) took {t1 - t0:.2f} seconds')
# run_algo(ds_january, ds_october, extents, grid1)
test_grid(grid1, ds_october, ds_january)

# t0 = time.time()
# grid2 = xr_to_grid2(ds_october, ds_january)
# t1 = time.time()
# print_grid_info(grid2, ds_october, ds_january)
# print(f'Grid 2 (Resolution) took {t1 - t0:.2f} seconds')
# # run_algo(ds_january, ds_october, extents, grid2)
# test_grid(grid2, ds_october, ds_january)

t0 = time.time()
grid3 = xr_to_grid3(ds_october, ds_january)
t1 = time.time()
print_grid_info(grid3, ds_october, ds_january)
print(f'Grid 3 (Numpy) took {t1 - t0:.2f} seconds')
# run_algo(ds_january, ds_october, extents, grid3)
test_grid(grid3, ds_october, ds_january)

print('done')


import pyscan
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
import tqdm
import matplotlib.patches as patches
import time

def xr_to_coords(ds):
    ds_stacked = ds.stack(xy=['x', 'y'])
    ds_non_nan = ds_stacked.dropna('xy')
    xs = ds.x.values
    x_non_nan = ds_non_nan.x.values
    ys = ds.y.values
    y_non_nan = ds_non_nan.y.values
    values_non_nan = ds_non_nan.value.values
    non_nan_array = np.column_stack((x_non_nan, y_non_nan, values_non_nan))
    return non_nan_array, np.unique(xs), np.unique(ys)

def plot_result(b_arr, m_arr, extents, rect=None, rect2=None):
    print('Plotting...')
    non_nan_b = b_arr[~np.isnan(b_arr)]
    non_nan_m = m_arr[~np.isnan(m_arr)]
    vmin = min(non_nan_b.min(), non_nan_m.min(), np.abs(non_nan_b - non_nan_m).min())
    vmax = max(non_nan_b.max(), non_nan_m.max(), np.abs(non_nan_b - non_nan_m).max())
    fig, axs = plt.subplots(1, 2, figsize=(6, 3))

    # Plot Baseline (b)
    im1 = axs[0].imshow(b_arr, cmap='coolwarm', vmin=vmin, vmax=vmax, extent=extents)
    axs[0].set_title('Baseline')

    # Plot Measure (m)
    im2 = axs[1].imshow(m_arr, cmap='coolwarm', vmin=vmin, vmax=vmax, extent=extents)
    axs[1].set_title('Measure')

    if rect is not None:
        for ax in axs:
            rect_patch = patches.Rectangle(
                (rect.lowX(), rect.lowY()), rect.upX() - rect.lowX(), rect.upY() - rect.lowY(),
                linewidth=2, edgecolor='k', facecolor='none'
            )
            ax.add_patch(rect_patch)
        axs[0].text(rect.lowX() -1, rect.lowY() - 1, 'Area Limited \n Subgrid \n (750 cells)', color='k', fontsize=11)

    if rect2 is not None:
        for ax in axs:
            rect_patch = patches.Rectangle(
                (rect2.lowX(), rect2.lowY()), rect2.upX() - rect2.lowX(), rect2.upY() - rect2.lowY(),
                linewidth=2, edgecolor='r', facecolor='none'
            )
            ax.add_patch(rect_patch)
        axs[0].text(rect2.lowX(), rect2.lowY() - 0.5, 'Unlimited Subgrid', color='darkred', fontsize=11)

    plt.tight_layout()
    plt.show()



ds_october = xr.open_dataset(f"data/pm_utah_2016_october.nc")
ds_january = xr.open_dataset(f"data/pm_utah_2016_january.nc")

max_slice = min(len(ds_october.x), len(ds_october.y))
# max_slice = 100
ds_m = ds_october.isel(x=slice(0, max_slice), y=slice(0, max_slice))
ds_b = ds_january.isel(x=slice(0, max_slice), y=slice(0, max_slice))
extents = [ds_october.x.min(), ds_october.x.max(), ds_october.y.min(), ds_october.y.max()]

m_arr, unique_xs, unique_ys = xr_to_coords(ds_m)
b_arr, _, _ = xr_to_coords(ds_b)
    
grid = pyscan.Grid(unique_xs, unique_ys, m_arr, b_arr)

disc_f = pyscan.RKULLDORF

max_area = 760 # given by the number of cells incldued

print('Limited grid size run...')
t0 = time.time()
subgrid = pyscan.max_subgrid_convex(grid, 1e-5, disc_f, max_area)
t1 = time.time()
print('Time in seconds:', (t1-t0))
rect = grid.toRectangle(subgrid)

cell_size_x = ds_october.x[1] - ds_october.x[0]
cell_size_y = ds_october.y[1] - ds_october.y[0]

col0, col1, row0, row1 = subgrid.lowCol(), subgrid.upCol(), subgrid.lowRow(), subgrid.upRow()

width = col1 - col0 + 1
height = row1 - row0 + 1
fvalue = subgrid.fValue()

xB, xU, yB, yU = rect.lowX(), rect.upX(), rect.lowY(), rect.upY()
# Rect size in number of cells
rect_size_area = width * height
print(f'Fvalue: {fvalue}')
print(f'Width: {width} cells, Height: {height} cells')
print(f'Area: {rect_size_area} cells')
print(f'Area limit: {max_area} cells')
print(f'Rectangle: ({col0}, {col1}), ({row0}, {row1})')

# REGULAR
disc_f = pyscan.RKULLDORF
print('Unlimited grid size run...')
t0 = time.time()
subgrid = pyscan.max_subgrid_convex(grid, 1e-5, disc_f)
t1 = time.time()
print('Time in seconds:', (t1-t0))
rect2 = grid.toRectangle(subgrid)

plot_result(ds_january.value.values, ds_october.value.values, extents, rect, rect2)
print('done')


import math
from collections import deque

def kadane_unrestricted(weights):
    print("=== Unrestricted Kadane ===")
    max_sum = float("-inf")
    curr_sum = 0
    start = 0
    best = (0, 0)

    for end, val in enumerate(weights):
        if curr_sum <= 0:
            curr_sum = val
            start = end
        else:
            curr_sum += val

        print(f"curr_sum={curr_sum}, start={start}, end={end}")
        if curr_sum > max_sum:
            max_sum = curr_sum
            best = (start, end)

    print(f"Best subarray: {best} with sum {max_sum}\n")
    return best, max_sum

def kadane_area_limited(weights, max_height):
    print("=== Area-Limited Kadane (post-check) ===")
    max_sum = float("-inf")
    curr_sum = 0
    start = 0
    best = (0, 0)

    for end, val in enumerate(weights):
        
        height = end - start + 1
        curr_sum += val
        
        if curr_sum <= 0:
            curr_sum = val
            start = end
            continue
            

        print(f"curr_sum={curr_sum}, height={height}, start={start}, end={end}")
        if height <= max_height and curr_sum > max_sum:
            max_sum = curr_sum
            best = (start, end)

    print(f"Best subarray: {best} with sum {max_sum}\n")
    return best, max_sum

def kadane_area_improved(weights, max_height):
    print("=== Improved Area-Limited Kadane (sliding window) ===")
    max_sum = float("-inf")
    curr_sum = 0
    start = 0
    best = (0, 0)

    for end, val in enumerate(weights):
        curr_sum += val
        height = end - start + 1

        while height > max_height:
            print(f"Shrinking: height={height} > {max_height}")
            curr_sum -= weights[start]
            start += 1
            height = end - start + 1

        print(f"curr_sum={curr_sum}, height={height}, start={start}, end={end}")
        if curr_sum > max_sum:
            max_sum = curr_sum
            best = (start, end)

    print(f"Best subarray: {best} with sum {max_sum}\n")
    return best, max_sum

def kadane_with_max_len(arr, L):
    """
    Max subarray sum with length <= L.
    Returns (best_sum, (start, end)).
    """

    n = len(arr)
    prefix_sum = [0] * (n + 1)  # size n+1
    for i in range(len(prefix_sum) - 1):  # 0 to n-1
        prefix_sum[i + 1] = prefix_sum[i] + arr[i]

    dq = deque([])
    max_sum = -math.inf
    best_seg = (0, 0)

    for i in range(0, len(prefix_sum)):

        print(f"i={i}, dq={list(dq)}")

        if dq and i - dq[0] > L:
            print(f"Removing {dq[0]} from deque because length exceeds {L}")
            dq.popleft()
        
        while dq and prefix_sum[dq[-1]] > prefix_sum[i]:
            print(f"Removing {dq[-1]} from deque because prefix_sum is larger than {prefix_sum[i]}")
            dq.pop()

        if dq and prefix_sum[i] - prefix_sum[dq[0]] > max_sum:
            print(f"New max found: {prefix_sum[i]} - {prefix_sum[dq[0]]} = {prefix_sum[i] - prefix_sum[dq[0]]}")
            max_sum = prefix_sum[i] - prefix_sum[dq[0]]
            best_seg = (dq[0], i - 1)
        
        print(f"Added {i} to deque, current deque: {list(dq)}")
        dq.append(i)

    print(f"Best subarray: {best_seg} with sum {max_sum}\n")
    return max_sum, best_seg


if __name__ == "__main__":
    weights = [3, -1, 5, -7, ]
    correct = 5
    max_height = 2

    b, m = kadane_unrestricted(weights)
    b, m = kadane_area_limited(weights, max_height)
    b, m = kadane_area_improved(weights, max_height)
    b, m = kadane_with_max_len(weights, max_height)



import pyscan
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
import tqdm
import matplotlib.patches as patches
import time

def plot_result(b_arr, m_arr, extents, rect=None, rect2=None, max_area=None):
    print('Plotting...')
    non_nan_b = b_arr[~np.isnan(b_arr)]
    non_nan_m = m_arr[~np.isnan(m_arr)]
    vmin = min(non_nan_b.min(), non_nan_m.min(), np.abs(non_nan_b - non_nan_m).min())
    vmax = max(non_nan_b.max(), non_nan_m.max(), np.abs(non_nan_b - non_nan_m).max())
    fig, axs = plt.subplots(1, 2, figsize=(6, 3))

    # Plot Baseline (b)
    im1 = axs[0].imshow(b_arr, cmap='coolwarm', vmin=vmin, vmax=vmax, extent=extents,origin='lower')
    axs[0].set_title('Baseline')

    # Plot Measure (m)
    im2 = axs[1].imshow(m_arr, cmap='coolwarm', vmin=vmin, vmax=vmax, extent=extents, origin='lower')
    axs[1].set_title('Measure')

    if rect is not None:
        for ax in axs:
            rect_patch = patches.Rectangle(
                (rect.lowX(), rect.lowY()), rect.upX() - rect.lowX(), rect.upY() - rect.lowY(),
                linewidth=2, edgecolor='k', facecolor='none'
            )
            ax.add_patch(rect_patch)
        axs[0].text(rect.lowX() -1, rect.lowY() - 1, f'Area Limited \n Subgrid \n ({max_area} cells)', color='k', fontsize=11)

    if rect2 is not None:
        for ax in axs:
            rect_patch = patches.Rectangle(
                (rect2.lowX(), rect2.lowY()), rect2.upX() - rect2.lowX(), rect2.upY() - rect2.lowY(),
                linewidth=2, edgecolor='r', facecolor='none'
            )
            ax.add_patch(rect_patch)
        axs[0].text(rect2.lowX(), rect2.lowY() - 0.5, 'Unlimited Subgrid', color='darkred', fontsize=11)

    plt.tight_layout()
    plt.show()

def grid_to_np(grid):
    print('Converting grid to numpy...')
    size = grid.size()
    m_arr = np.zeros((size, size))
    b_arr = np.zeros((size, size))
    for y in tqdm.tqdm(range(size)):
        for x in range(size):
            b_arr[y, x] = grid.blueWeight(y, x)  # Note: 
            m_arr[y, x] = grid.redWeight(y, x)
    # remember, in pyscan grid, y,x are sorted from small to large. So we have to flip the y-axis
    # m_arr = np.flipud(m_arr)
    # b_arr = np.flipud(b_arr)
    return m_arr, b_arr


unique_xs = np.array([0, 1, 2])
unique_ys = np.array([0, 1, 2])
extents = (unique_xs[0], unique_xs[-1], unique_ys[0], unique_ys[-1])
b_arr = np.ones((len(unique_ys), len(unique_xs)))
m_arr = np.ones((len(unique_ys), len(unique_xs)))
# Add a block of 3x3 1s in m_arr
m_arr[1:3, 1:3] = 2

print("Baseline array (b_arr):")
print(b_arr)
print("Measure array (m_arr):")
print(m_arr)

# transform m_arr to N,3 2D array of coord, coord, value
n_m_arr = np.zeros((m_arr.size, 3))
n_b_arr = np.zeros((b_arr.size, 3))
coords = np.array(np.meshgrid(unique_xs, unique_ys)).T.reshape(-1, 2)
n_m_arr[:, :2] = coords
n_b_arr[:, :2] = coords
for i in range(coords.shape[0]):
    n_m_arr[i, 2] = m_arr[coords[i, 1], coords[i, 0]]
    n_b_arr[i, 2] = b_arr[coords[i, 1], coords[i, 0]]
m_arr = n_m_arr
b_arr = n_b_arr

grid = pyscan.Grid(unique_xs, unique_ys, m_arr, b_arr)

m_arr, b_arr = grid_to_np(grid)
print("Converted Measure array (m_arr):")
print(m_arr)
print("Converted Baseline array (b_arr):")
print(b_arr)


disc_f = pyscan.RKULLDORF

max_area = 1000

print('Limited grid size run...')
subgrid = pyscan.max_subgrid_linear(grid, 1, -1, max_area)

rect = grid.toRectangle(subgrid)
col0, col1, row0, row1 = subgrid.lowCol(), subgrid.upCol(), subgrid.lowRow(), subgrid.upRow()

width = col1 - col0
height = row1 - row0
fvalue = subgrid.fValue()

cell_height_x = unique_xs[1] - unique_xs[0]
cell_width_y = unique_ys[1] - unique_ys[0]

xB, xU, yB, yU = rect.lowX(), rect.upX(), rect.lowY(), rect.upY()

# Rect size in number of cells
rect_size_area = width * height
print(f'Fvalue: {fvalue}')
print(f'Width: {width} cells, Height: {height} cells')
print(f'Area: {rect_size_area} cells')
print(f'Area limit: {max_area} cells')
print(f'Rectangle: ({col0}, {col1}), ({row0}, {row1})')
print(f"Rectangle coordinates: ({xB}, {yB}), ({xU}, {yU})")

plot_result(b_arr, m_arr, extents, rect, max_area=rect_size_area)

print('done')


from .libpyscan import *
from . import libpyscan as lp
import random
import itertools
import math



def bounding_box(pts):
    if not pts:
        return None
    min_x = pts[0][0]
    max_x = pts[0][0]
    min_y = pts[0][1]
    max_y = pts[0][1]
    for pt in pts:
        min_x = min(min_x, pt[0])
        max_x = max(max_x, pt[0])
        min_y = min(min_y, pt[1])
        max_y = max(max_y, pt[1])
    return ((min_x, min_y), (max_x, max_y))


def to_weighted(points):
    return [WPoint(1.0, pt[0], pt[1], 1.0) for pt in points]


def trajectories_to_flux(trajectories):
    return [t[0] for t in trajectories], [t[-1] for t in trajectories]


def trajectories_to_labels(trajectories):
    return itertools.chain.from_iterable([[LPoint(label, 1.0, pt[0], pt[1], 1.0) for pt in traj]
            for label, traj in zip(range(len(trajectories)), trajectories)])


def my_sample(samp, count):
    return random.sample(samp, min(len(samp), int(count + .5)))


def evaluate_range(range, mp, bp, disc_f):
    """
    Evaluates this range to compute the total discrepancy over a set of points.

    :param range: Some arbitrary range.
    :param mp: measured set
    :param bp: baseline set
    :param disc_f: Discrepancy function.
    :return: Discrepancy function value.
    """
    if not mp and not bp:
        return evaluate(disc_f, 0, 0, 0, 0)
    elif not mp:
        pt_obj = bp[0]
    else:
        pt_obj = mp[0]

    if isinstance(pt_obj, LPoint):
        if isinstance(range, Disk):
            return evaluate_disk_labeled(range, mp, bp, disc_f)
        elif isinstance(range, Halfplane):
            return evaluate_disk_labeled(range, mp, bp, disc_f)
        elif isinstance(range, Rectangle):
            return evaluate_rectangle_labeled(range, mp, bp, disc_f)
    elif isinstance(pt_obj, WPoint):
        if isinstance(range, Disk):
            return evaluate_disk(range, mp, bp, disc_f)
        elif isinstance(range, Halfplane):
            return evaluate_halfplane(range, mp, bp, disc_f)
        elif isinstance(range, Rectangle):
            return evaluate_rectangle(range, mp, bp, disc_f)
    raise ValueError()


def evaluate_range_trajectory(range, mp, bp, disc_f):
    """
    Evaluates this range to compute the total discrepancy over a set of trajectories.

    :param range: Some arbitrary range.
    :param mp: measured set
    :param bp: baseline set
    :param disc_f: Discrepancy function.
    :return: Discrepancy function value.
    """

    if not mp and not bp:
        return evaluate(disc_f, 0, 0, 0, 0)
    if isinstance(range, Disk):
        return evaluate_disk_trajectory(range, mp, bp, disc_f)
    elif isinstance(range, Halfplane):
        return evaluate_halfplane_trajectory(range, mp, bp, disc_f)
    elif isinstance(range, Rectangle):
        return evaluate_rectangle_trajectory(range, mp, bp, disc_f)
    else:
        raise ValueError()


def split_set(pts, rate):
    red_set = my_sample(pts, len(pts) * rate)
    red_set_set = set(red_set)
    return red_set, [item for item in pts if item not in red_set_set]


def plant_region(points, r, p, q, eps, scan_f):
    """
    This takes a scanning function and two point sets and then computes a planted region that contains some fraction r
    of the points with some tolerance.
    This then computes a red and blue set of points based on the planted region.

    Inside the region q fraction of points are red.
    Outside the region p fraction of points are red

    :param points: List of points.
    :param r: Fraction of points inside the region.
    :param p: Fraction of points outside the region that are red.
    :param q: Fraction of points inside the region that are red.
    :param eps: Difference between r and the fraction of points inside the region.
    :param scan_f: The scan function to use (ex max_disk)
    :return: red set, blue set, region planted.
    """
    while True:

        net_size = int(1 / min(r, eps) + 1)
        s_size = int(1 / min(r, eps)** 2 + 1)

        net = my_sample(points, net_size)
        sample = my_sample(points, s_size)
        disc = size_region(r)
        reg, _ = scan_f(net, sample, [], disc)

        val = evaluate_range(reg, points, [], disc)
        if 1 - val < eps:
            break

    in_region = []
    out_region = []
    for pt in pts:
        if reg.contains(pt):
            in_region.append(pt)
        else:
            out_region.append(pt)

    red_in, blue_in = split_set(in_region, q)

    red_out, blue_out = split_set(out_region, p)
    return red_in + red_out, blue_in + blue_out, reg


def plant_full_square(trajectories, r, p, q, max_count=32):
    """
    Choose a point at random from a trajectory and then expand a square out from this point till this region contains
    r fraction of all the trajectories.

    :param trajectories: List of list of points.
    :param r: double between 0 and 1
    :param p: double between 0 and 1
    :param q: double between 0 and 1
    :param disc: The discrepancy function to evaluate exactly on this region.
    :param max_count: The maximum number of times we will attempt to find the right sized region.
    :return: red set of trajectories, blue set of trajectories, the planted region, and the exact partial discrepancy.
    """
    if not trajectories:
        return None
    count = 0
    traj = []
    while not traj:
        count+= 1
        traj = random.choice(trajectories)
        if count > 10 :
            return None

    seed_pt = random.choice(traj)
    upper_bound = 1.0
    lower_bound = 0.0
    num = 0
    while num < max_count:
        size = (upper_bound + lower_bound) / 2
        reg = Rectangle(seed_pt[0] + size / 2, seed_pt[1] + size / 2, seed_pt[0] - size / 2, seed_pt[1] - size / 2)
        count = sum(1 for traj in trajectories if reg.intersects_trajectory(Trajectory(traj)))
        if abs(count - r * len(trajectories)) <= 2:
            break
        if count - r * len(trajectories) > 0:
            upper_bound = size
        else:
            lower_bound = size
        num += 1


    inside_rect = [traj for traj in trajectories if reg.intersects_trajectory(Trajectory(traj))]
    red_in, blue_in = split_set([tuple(traj) for traj in inside_rect], q)
    outside_rect = [traj for traj in trajectories if not reg.intersects_trajectory(Trajectory(traj))]
    red_out, blue_out = split_set([tuple(traj) for traj in outside_rect], p)

    diff = evaluate(disc, len(red_in), len(red_in) + len(red_out), len(blue_in), len(blue_in) + len(blue_out))
    return red_in + red_out, blue_in + blue_out, reg



def plant_full_halfplane(trajectories, r, p, q):
    """
    Choose a random direction and then finds a halfplane with this normal containing r fraction of the total trajectories.

    :param trajectories: List of list of points.
    :param r: Fraction of trajectories in region.
    :param p: Fraction of red trajectories outside of region.
    :param q: Fraction of red trajectories inside of region.
    :param disc: Discrepancy function to measure exactly on region.
    :return: red set of trajectories, blue set of trajectories, planted region, discrepancy function value.
    """
    def min_distance(pts, direc):
        return min([direc[0] * pt[0] + direc[1] * pt[1] for pt in pts])
    rand_direc = (random.gauss(0,1), random.gauss(0, 1))
    trajectories = sorted(trajectories, key=lambda x: min_distance(x, rand_direc))

    pt = min(trajectories[int(r * len(trajectories))], key=lambda pt: rand_direc[0] * pt[0] + rand_direc[1] * pt[1])
    lc = pt[0] * rand_direc[0] + pt[1] * rand_direc[1]
    plant_region = Halfplane(Point(rand_direc[0], rand_direc[1], -lc))

    inside_plane = trajectories[:int(r * len(trajectories))]
    outside_plane = trajectories[int(r * len(trajectories)):]
    red_in, blue_in = split_set([tuple(traj) for traj in inside_plane], q)
    red_out, blue_out = split_set([tuple(traj) for traj in outside_plane], p)

    return red_in + red_out, blue_in + blue_out, plant_region


def plant_partial_halfplane(trajectories, r, p, q, eps):
    """
    Choose a random direction and expand the region along this direction till we contain r fraction of the total
    trajectory arc length.

    :param trajectories: List of list of points.
    :param r: double between 0 and 1
    :param p: double between 0 and 1
    :param q: double between 0 and 1
    :param eps: This defines the maximum difference between r and the fraction of the trajectories that are in the found region.
    :param disc: The discrepancy function to evaluate exactly on this region.
    :return: red set of trajectories, blue set of trajectories, the planted region, and the exact partial discrepancy.
    """

    def min_distance(pt, direc):
        return direc[0] * pt[0] + direc[1] * pt[1]
    #so it always points down
    rand_direc = (random.gauss(0,1), -abs(random.gauss(0, 1)))

    trajectory_obj = [Trajectory(pts) for pts in trajectories]
    all_pts = uniform_sample(trajectory_obj, int(1 / eps ** 2 + 1), False)
    all_pts = sorted(all_pts, key=lambda x: min_distance(x, rand_direc))

    pt = all_pts[int((1 - r) * len(all_pts))]
    lc = pt[0] * rand_direc[0] + pt[1] * rand_direc[1]
    plant_region = Halfplane(Point(rand_direc[0], rand_direc[1], -lc))

    inside_disk = [traj for traj in trajectory_obj if plant_region.intersects_trajectory(traj)]
    outside_disk = [traj for traj in trajectory_obj if not plant_region.intersects_trajectory(traj)]

    red_in, blue_in = split_set(inside_disk, q)
    red_out, blue_out = split_set(outside_disk, p)
    return red_in + red_out, blue_in + blue_out, plant_region


def plant_full_disk(trajectories, r, p, q):
    """
    Choose a point at random from a trajectory and then expand outward from there until we find a disk contains r fraction
    of the trajectories.

    :param trajectories: List of trajectories. This can be rearanged.
    :param r: Fraction of trajectories in region.
    :param p: Fraction of red trajectories outside of region.
    :param q: Fraction of red trajectories inside of region.
    :param disc: Discrepancy function to measure exactly on region.
    :return: red set, blue set, planted disk, maximum discrepancy.
    """
    traj = trajectories[random.randint(0, len(trajectories)-1)]
    origin = traj[random.randint(0, len(traj)-1)]

    trajectories.sort(key=lambda el: Trajectory(el).point_dist(origin))

    trajectories_ix = list(range(len(trajectories)))

    inside_disk = trajectories_ix[:int(r * len(trajectories))]
    red_in, blue_in = split_set(inside_disk, q)
    max_disk = Disk(origin[0], origin[1], Trajectory(trajectories[inside_disk[-1]]).point_dist(origin))
    del inside_disk

    outside_disk = trajectories_ix[int(r * len(trajectories)):]
    red_out, blue_out = split_set(outside_disk, p)
    del outside_disk

    red_ix = red_in + red_out
    blue_ix = blue_in + blue_out

    #sort the trajectories with respect to the ix.
    return [trajectories[i] for i in red_ix], [trajectories[i] for i in blue_ix], max_disk


def plant_partial_disk(trajectories, r, p, q, eps):
    """
    Choose a point at random from a trajectory and then expand outward from there. Computes the fraction of length inside the disk
    for each segment and then does bisection on this amount since it is a monotonic function to compute within eps fraction.

    :param trajectories: List of list of points.
    :param r: double between 0 and 1
    :param p: double between 0 and 1
    :param q: double between 0 and 1
    :param eps: This defines the maximum difference between r and the fraction of the trajectories that are in the found region.
    :param disc: The discrepancy function to evaluate exactly on this region.
    :return: red set of trajectories, blue set of trajectories, the planted region, and the exact partial discrepancy.
    """

    origin_tuple = random.choice(list(itertools.chain.from_iterable(trajectories)))
    origin = Point(origin_tuple[0], origin_tuple[1], 1.0)
    trajectory_obj = [Trajectory(pts) for pts in trajectories]
    all_pts = uniform_sample(trajectory_obj, int(1 / eps ** 2 + 1), False)
    trajectories = sorted(all_pts, key=lambda x: x.dist(origin))

    disk_boundary = trajectories[int(r * len(all_pts))]
    max_disk = Disk(origin[0], origin[1], origin.dist(disk_boundary))
    inside_disk = [traj for traj in trajectory_obj if max_disk.intersects_trajectory(traj)]
    outside_disk = [traj for traj in trajectory_obj if not max_disk.intersects_trajectory(traj)]

    red_in, blue_in = split_set(inside_disk, q)
    red_out, blue_out = split_set(outside_disk, p)
    return red_in + red_out, blue_in + blue_out, max_disk


def distribution(points, p, scan_f, n, s, disc=DISC):
    while True:
        red, blue = split_set(points, p)

        net_set = my_sample(red, min(len(red), n)) + my_sample(blue, min(len(blue), n))
        m_sample = my_sample(red, min(len(red), s), red)
        b_sample = my_sample(blue, min(len(blue), s), blue)
        reg, val = scan_f(net_set, m_sample, b_sample, disc)

        yield val


def null_cdf(observations):
    values = sorted(observations)
    prob = [x / float(len(observations)) for x in range(len(observations))]
    return values, prob


def random_rect(points, r):
    """
    Plants a random rectangle containing r fraction of the points.

    :param points: List of points
    :param r: Fraction of points inside of planted region.
    :return: The planted region.
    """
    def bBox(*bb):
        return (min(*map(lambda pt: pt[0], bb)),
                max(*map(lambda pt: pt[0], bb)),
                min(*map(lambda pt: pt[1], bb)),
                max(*map(lambda pt: pt[1], bb)))

    while True:
        [pt1, pt2] = random.sample(points, 2)

        if random.randint(0, 1) == 0:
            (minX, maxX, minY, maxY) = bBox(pt1, pt2)
            fx = lambda pt: pt[0]
            fy = lambda pt: pt[0]
        else:
            fx = lambda pt: pt[0]
            fy = lambda pt: pt[0]
            (minY, maxY, minX, maxX) = bBox(pt1, pt2)

        if random.randint(0, 1) == 0:
            sf = lambda pt: fx(pt)
        else:
            sf = lambda pt: -fx(pt)
            minX, maxX = maxX, minX

        subPoints = [pt for pt in points if minY <= fy(pt) and fy(pt) <= maxY]
        # Make a pass through the data and find the lower and upper bound.
        inB = [pt for pt in subPoints if minX <= fx(pt) and fx(pt) <= maxX]
        if random.randint(0, 1):  # random.randint(0, 1) == 0:
            ab = [pt for pt in subPoints if maxX < fx(pt)]
            ab.sort(key=lambda p: -sf(p))
        else:
            ab = [pt for pt in subPoints if fx(pt) < minX]
            ab.sort(key=lambda p: sf(p))
        while len(ab) > 0:
            if len(inB) > int(r * len(points) + .5):
                break
            if len(inB) == int(r * len(points) + .5):
                (lx, ux, ly, uy) = bBox(*inB)
                return Rectangle(ux, uy, lx, ly)
            el = ab.pop()
            inB.append(el)
            while len(ab) > 0 and sf(el) == sf(ab[-1]):
                el = ab.pop()
                inB.append(el)


def plant_rectangle(pts, r, p, q):
    """
    Create a set of red and blue points with a random rectangle planted containing r fraction of the points with
    q fraction of the points in the region being red and p fraction of the points outside of the region being red.

    :param pts: List of points.
    :param r: Fraction of points contained in the planted region
    :param p: Fraction of points outside region that red.
    :param q: Fraction of points inside region that are red.
    :return: red set, blue set, planted region.
    """

    rect = random_rect(pts, r)

    inside_rect = []
    outside_rect = []
    for pt in pts:
        if rect.contains(pt):
            inside_rect.append(pt)
        else:
            outside_rect.append(pt)

    red_in, blue_in = split_set(inside_rect, q)
    red_out, blue_out = split_set(outside_rect, p)
    return red_in + red_out, blue_in + blue_out, rect


def plant_disk(pts, r, p, q):
    """
    Create a set of red and blue points with a random disk planted containing r fraction of the points with
    q fraction of the points in the region being red and p fraction of the points outside of the region being red.

    :param pts: List of points.
    :param r: Fraction of points contained in the planted region
    :param p: Fraction of points outside region that red.
    :param q: Fraction of points inside region that are red.
    :return: red set, blue set, planted region.
    """

    selected = my_sample(pts, 1)
    if not selected:
        return None
    origin = selected[0]

    ordered = sorted(pts, key=lambda x: x.dist(origin))
    disk_boundary = ordered[int(r * len(ordered))]
    max_disk = Disk(origin[0], origin[1], origin.dist(disk_boundary))

    inside_disk = ordered[:int(r * len(ordered))]
    outside_disk = ordered[int(r * len(ordered)):]

    red_in, blue_in = split_set(inside_disk, q)
    red_out, blue_out = split_set(outside_disk, p)
    return red_in + red_out, blue_in + blue_out, max_disk


def plant_halfplane(pts, r, p, q):
    """
    Create a set of red and blue points with a random halfplane planted containing r fraction of the points with
    q fraction of the points in the region being red and p fraction of the points outside of the region being red.

    :param pts: List of points
    :param r: Fraction of points contained in the planted region
    :param p: Fraction of points outside region that red.
    :param q: Fraction of points inside region that are red.
    :return: red set, blue set, planted region.
    """

    def min_distance(pt, direc):
        return direc[0] * pt[0] + direc[1] * pt[1] 
    rand_direc = (random.gauss(0, 1), abs(random.gauss(0, 1)))

    ordered = sorted(pts, key=lambda x: min_distance(x, rand_direc))
    inside_halfplane = ordered[:int(r * len(ordered))]
    outside_halfplane = ordered[int(r * len(ordered)):]

    pt = ordered[int(r * len(ordered))]
    lc = pt[0] * rand_direc[0] + pt[1] * rand_direc[1]


    plant_region = Halfplane(Point(-rand_direc[0], -rand_direc[1], lc))

    red_in, blue_in = split_set(inside_halfplane, q)
    red_out, blue_out = split_set(outside_halfplane, p)
    return red_in + red_out, blue_in + blue_out, plant_region


def plant_partial_rectangle(trajectories, r, p, q, eps):
    """
    This plants a region containing r fraction of the total arc length of all trajectories. q fraction of the trajectories
    crossing this region are assigned to be anomalous and p fraction of trajectories not crossing this region are assigned
    to be anomalous.

    :param trajectories: List of list of points.
    :param r: double between 0 and 1
    :param p: double between 0 and 1
    :param q: double between 0 and 1
    :param eps: This defines the maximum difference between r and the fraction of the trajectories that are in the found region.
    :return: red set of trajectories, blue set of trajectories, the planted region.
    """
    trajectory_obj = [Trajectory(pts) for pts in trajectories]
    all_pts = uniform_sample(trajectory_obj, int(1 / eps ** 2 + 1), False)
    _, _, rect = plant_rectangle(all_pts, r, p, q)
    inside_rect = [traj for traj in trajectory_obj if rect.intersects_trajectory(traj)]
    outside_rect = [traj for traj in trajectory_obj if not rect.intersects_trajectory(traj)]
    red_in, blue_in = split_set(inside_rect, q)
    red_out, blue_out = split_set(outside_rect, p)

    return red_in + red_out, blue_in + blue_out, rect




def paired_plant_region(traj_start, traj_end, r, q, region_plant_f):
    """
    This plants a region where every trajectory
    completely outside or inside of the region has an endpoint chosen at random.
    Every trajectory with one endpoint inside the region has an endpoint chosen inside
    with probability q (exactly q fraction have one endpoint in the region)
    traj_start and traj_end should be the same length.

    :param traj_start: List of points.
    :param traj_end: List of points.
    :param r: Fraction of points in the region
    :param q: Fraction of points in the region that are anomalous
    :param region_plant_f: Scanning function to use to find the region (example max_disk)
    :return: Red planted set, blue planted set, and the planted region.
    """
    _, _, reg = region_plant_f(traj_start + traj_end, r, .5, q)

    flux_region = []
    out_region = []
    for st_pt, end_pt in zip(traj_start, traj_end):
        if reg.contains(st_pt) and not reg.contains(end_pt):
            flux_region.append((st_pt, end_pt))
        elif reg.contains(end_pt) and not reg.contains(st_pt):
            flux_region.append((end_pt, st_pt))
        else:
            if random.random() <= .5:
                out_region.append((st_pt, end_pt))
            else:
                out_region.append((end_pt, st_pt))


    q_fraction, remainder = split_set(flux_region, q)
    remainder = [(ep, sp) for (sp, ep) in remainder]

    red, blue = zip(*(q_fraction + remainder + out_region))
    return red, blue, reg



def plant_kernel_disk_region(pts, r, p, q, eps=.0001):
    """
    Plants a region using the bernoulli model with the given p and q parameters.
    Need the fraction in the q set to be r.

    Returns the bandwidth of the anomaly.
    :param pts:
    :param r:
    :param p:
    :param q:
    :param eps: The error on the planted r parameter.
    :return:
    """
    if not pts:
        return None
    ((min_x, min_y), (max_x, max_y)) = bounding_box(pts)


    def sq_dist(x, pt):
        return (x[0] - pt[0])**2 + (x[1] - pt[1])**2

    [p1, p2] = random.sample(pts, 2)
    alpha = random.random()
    px = p1[0] * alpha + p2[0] * (1 - alpha)
    py = p1[1] * alpha + p2[1] * (1 - alpha)

    #Find the correct bandwidth by doing bisection
    seeds = [random.random() for p in pts]
    bandwidth_upper = 2 * max((max_x - min_x), (max_y - min_y))
    bandwidth_lower = 0.0

    while True:
        bandwidth = (bandwidth_upper + bandwidth_lower) / 2
        def kernel(x):
            return math.exp(-sq_dist(x, (px, py)) / (2 * bandwidth**2))

        p_count = 0
        for i, pt in enumerate(pts):
            if seeds[i] < kernel(pt):
                p_count += 1

        if r - eps <= p_count / len(pts) <= r + eps:
            break
        if r + eps < p_count / len(pts):
            bandwidth_upper = bandwidth
        else:
            bandwidth_lower = bandwidth

    measured_set = []
    baseline_set = []
    p_count = 0
    q_count = 0
    for i, pt in enumerate(pts):
        if seeds[i] < kernel(pt):
            p_count += 1
            if random.random() < p:
                measured_set.append(pt)
            else:
                baseline_set.append(pt)
        else:
            q_count += 1
            if random.random() < q:
                measured_set.append(pt)
            else:
                baseline_set.append(pt)

    return measured_set, baseline_set, bandwidth, (px, py)



def disc_bernoulli_kern(measured, baseline, p, q, bandwidth, center):
    def sq_dist(x, pt):
        return (x[0] - pt[0])**2 + (x[1] - pt[1])**2
    def kernel(x):
        return math.exp(-sq_dist(x, center) / bandwidth**2)

    act_disc = 0
    for pt in measured:
        fr = kernel(pt)
        gr = p * fr + (1 - fr) * q
        act_disc += math.log(gr)
    for pt in baseline:
        fr = kernel(pt)
        gr = p * fr + (1 - fr) * q
        act_disc += math.log(1 - gr)

    scale = len(measured) / (len(measured) + len(baseline))
    null_val = len(measured) * math.log(scale) + len(baseline) * math.log(1 - scale)
    return act_disc - null_val



def close_region(region):
    """
    Ensures that this region starts and ends at the same point. This is useful for using trajectory methods
    for regions.
    :param region: List of points
    :return: list of points
    """
    if region:
        if Point(region[-1][0], region[-1][1], 1.0).approx_eq(region[0]):
            region.append(region[0])

def plant_full_disk_region(region_set, r, p, q):
    """
    Choose a point at random from a region and then expand a disk out from this point till this disk contains
    r fraction of all the regions.
    :param region_set: List of list of points. This argument is modified by adding a point at the end in some cases.
    :param r: double between 0 and 1
    :param p: double between 0 and 1
    :param q: double between 0 and 1
    :return: red set of regions, blue set of regions, the planted region
    """
    for reg in region_set:
        close_region(reg)
    return plant_full_disk(region_set, r, p, q)


try:
    from shapely.geometry import Polygon
    from shapely import geometry



    def measure_range_region(reg_geom, range_set, weights=None):
        weight_sum = 0
        if weights is None:
            weights = [1.0] * len(range_set)
        if isinstance(reg_geom, Disk):

            for i in range(len(range_set)):
                poly = Polygon(range_set[i])
                pt = reg_geom.get_origin()
                pt_tpl = geometry.Point(pt[0], pt[1])
                if poly.distance(pt_tpl) <= reg_geom.get_radius():
                    weight_sum += weights[i]
        elif isinstance(reg_geom, Halfplane):
            for i in range(len(range_set)):
                traj = Trajectory(range_set[i])
                if reg_geom.intersects_trajectory(traj):
                    weight_sum += weights[i]

        elif isinstance(reg_geom, Rectangle):
            rect = geometry.box(reg_geom.lowX(), reg_geom.lowY(), reg_geom.upX(), reg_geom.upY())
            for i in range(len(range_set)):
                poly = Polygon(range_set[i])
                if rect.intersects(poly):
                    weight_sum += weights[i]
        else:
            raise ValueError()
        return weight_sum
except:
    pass

try:
    import matplotlib.pyplot as plt
    import numpy as np

    def plot_points(ax, pts, c):
        xs = []
        ys = []
        for pt in pts:
            xs.append(pt[0] )
            ys.append(pt[1])
        ax.scatter(xs, ys, color=c, marker='.')

    def plot_kernel(ax, pts, pt, bandwidth, res=20, transform=None):
        (mnx, mny), (mxx, mxy) = bounding_box(pts)
        mxx = np.linspace(mnx, mxx, res)
        mxy = np.linspace(mny, mxy, res)
        xv, yv = np.meshgrid(mxx, mxy)

        def kernel(x, y):
            return np.exp(-(np.power(x - pt[0], 2.0) + np.power(y - pt[1], 2.0) ) / bandwidth**2)
        ax.contourf(xv, yv, kernel(xv, yv), alpha=.3, antialiased=True, transform=transform, cmap="Blues")
        ax.contour(xv, yv, kernel(xv, yv), linewidths=4.0, antialiased=True, transform=transform)

except:
    pass

def plant_full_square_region(regions, r, p, q, max_count=32):
    """
    Choose a point at random from a region and then expand a square out from this point till this region contains
    r fraction of all the regions.

    :param regions: List of list of points. These lists are modified. In some cases the regions will have a point appended at the end to close them.
    :param r: double between 0 and 1
    :param p: double between 0 and 1
    :param q: double between 0 and 1
    :param max_count: The maximum number of times we will attempt to find the right sized region.
    :return: red set of regions, blue set of regions, the planted region
    """
    for reg in regions:
        close_region(reg)
    return plant_full_square(regions, r, p, q, max_count)

def max_disk_trajectory(net, red_sample, blue_sample, min_disk_r, max_disk_r, alpha, disc, fast_disk=True):
    """
    Computes the highest discrepancy disk over a set of trajectories. Executes at multiple scales using the grid
    directional compression method and internally compresses the trajectories if fast_disk is enabled.

    :param net: A list of trajectories (scaled to be in a 0 by 1 box).
    :param red_sample: A list of trajectories (scaled to be in a 0 by 1 box).
    :param blue_sample: A list of trajectories (scaled to be in a 0 by 1 box).
    :param min_disk_r: The minimum disk radius to consider.
    :param max_disk_r: The maximum disk radius to consider.
    :param alpha: The spatial error with which to approximate the trajectories.
    :param disc: The discrepancy function to use.
    :param fast_disk: Default True.
    :return:
    """
    mx = -1
    curr_disk_r = max(min_disk_r, alpha)
    reg = None
    while True:

        chord_l = math.sqrt(4 * alpha * curr_disk_r - 2 * alpha * alpha)
        m_sample = [grid_direc_kernel(dp_compress(traj, alpha), chord_l, alpha) for traj in red_sample]
        b_sample = [grid_direc_kernel(dp_compress(traj, alpha), chord_l, alpha) for traj in blue_sample]
        pt_net = [grid_direc_kernel(dp_compress(traj, alpha), chord_l, alpha) for traj in net]
        m_sample = list(trajectories_to_labels(m_sample))
        b_sample = list(trajectories_to_labels(b_sample))
        net_set = list(trajectories_to_labels(pt_net))


        new_reg, new_mx = max_disk_scale_labeled(net_set, m_sample, b_sample, fast_disk, curr_disk_r, disc)
        if new_mx > mx:
            reg = new_reg
            mx = new_mx
        curr_disk_r *= 2
        if curr_disk_r >= max_disk_r:
            break
    return reg, mx


def max_disk_trajectory_fixed(net, m_sample, b_sample, min_disk_r, max_disk_r,  disc, fast_disk=True):
    """
    Computes the highest discrepancy disk over a set of trajectories. Executes at multiple scales, but uses whatever set
    of points the trajectories have been compressed with.

    :param net: A list of trajectories (scaled to be in a 0 by 1 box).
    :param m_sample: A list of trajectories (scaled to be in a 0 by 1 box).
    :param b_sample: A list of trajectories (scaled to be in a 0 by 1 box).
    :param min_disk_r: The minimum disk radius to consider.
    :param max_disk_r: The maximum disk radius to consider.
    :param disc: The discrepancy function to use.
    :param fast_disk: Default True.
    :return: A tuple of the maximum disk and the corresponding maximum value.
    """
    mx = -1
    curr_disk_r = min_disk_r
    reg = None
    while curr_disk_r < max_disk_r:
        new_reg, new_mx = max_disk_scale_labeled(net, m_sample, b_sample, fast_disk, curr_disk_r, disc)
        if new_mx > mx:
            reg = new_reg
            mx = new_mx
        curr_disk_r *= 2
    return reg, mx


def max_disk_region(net, red_sample, red_weight, blue_sample, blue_weight, min_disk_r, max_disk_r, alpha, disc, fast_disk=True):
    """
    Computes the highest discrepancy disk over a set of trajectories. Executes at multiple scales using the grid
    directional compression method and internally compresses the trajectories if fast_disk is enabled.

    :param net: A list of trajectories
    :param red_sample: A list of trajectories
    :param blue_sample: A list of trajectories
    :param min_disk_r: The minimum disk radius to consider.
    :param max_disk_r: The maximum disk radius to consider.
    :param alpha: The spatial error with which to approximate the trajectories.
    :param disc: The discrepancy function to use.
    :param fast_disk: Default True.
    :return:
    """
    mx = -1
    curr_disk_r = max(min_disk_r, alpha)
    reg = None
    while True:

        m_sample = [polygon_grid_hull(reg, alpha, curr_disk_r) for reg in red_sample]
        b_sample = [polygon_grid_hull(reg, alpha, curr_disk_r) for reg in  blue_sample]
        pt_net = [polygon_grid_hull(reg, alpha, curr_disk_r) for reg in net]

        m_sample = [[LPoint(i, w, pt[0], pt[1], 1.0) for pt in reg] for reg, w, i in zip(m_sample, red_weight, range(len(m_sample)))]
        b_sample = [[LPoint(i, w, pt[0], pt[1], 1.0) for pt in reg] for reg, w, i in zip(b_sample, blue_weight, range(len(m_sample)))]
        m_sample = list(itertools.chain.from_iterable(m_sample))
        b_sample = list(itertools.chain.from_iterable(b_sample))
        net_set = list(trajectories_to_labels(pt_net))
        new_reg, new_mx = max_disk_scale_labeled(net_set, m_sample, b_sample, fast_disk, curr_disk_r, disc)
        if new_mx > mx:
            reg = new_reg
            mx = new_mx
        curr_disk_r *= 2
        if curr_disk_r >= max_disk_r:
            break
    return reg, mx


